/* 
Subtitles Workgroup: A subtitles creation software with network collaboration suppport
Copyright (C) 2007 smalleye

This program is free software; you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace Nikse.SubtitleEdit.Logic.Networking
{
    public class NetworkConnection
    {
        private class LogEntry
        {
            public NetworkOperation Operation;
            public NetworkOperation Response;
            public EventWaitHandle Locking;

            public LogEntry(NetworkOperation newOperation)
            {
                Operation = newOperation;
                Locking = new EventWaitHandle(false, EventResetMode.AutoReset);
                Response = null;
            }
        }

        private class OperationLog : ArrayList
        {
            public LogEntry Add(NetworkOperation operation)
            {
                LogEntry newentry = new LogEntry(operation);
                base.Add(newentry);
                return newentry;
            }

            public bool IsWaiting(NetworkOperation operation)
            {
                long wtrID = operation.trID;
                bool found = false;
                int c = 0;

                while (!found && c < Count)
                {
                    LogEntry entry = this[c];
                    found = entry.Operation.trID == wtrID;
                    c++;
                }

                return found;
            }

            public void SendSignal(NetworkOperation operation, bool delete)
            {
                long wtrID = operation.trID;
                bool found = false;
                int c = 0;

                while (!found && c < Count)
                {
                    LogEntry entry = this[c];
                    found = entry.Operation.trID == wtrID;
                    c++;
                }

                if (found)
                {
                    LogEntry entry = this[c - 1];
                    entry.Response = operation;
                    entry.Locking.Set();
                    if (delete)
                        base.Remove(entry);
                }
            }

            public new LogEntry this[int indice]
            {
                get
                {
                    return (LogEntry)base[indice];
                }
            }
            
        }

        
        public delegate void dlgClientDisconnect(int id);
        public delegate void dlgReceive(int id, NetworkOperation op);


        public TcpClient Client;
        private int Id;
        private System.Threading.Timer Clock;
        public int ConnectionId
        {
            get
            {
                return Id;
            }
            set
            {
                Id = value;
            }
        }

        private long _trId;
        private Thread _receiveTh;
        private OperationLog _log;

        public event dlgClientDisconnect OnClientDisconnect;
        public event dlgReceive OnReceive;

        public NetworkConnection(TcpClient newClient, int newId)
        {
            Client = newClient;
            Id = newId;
            _trId = 1;
            _log = new OperationLog();
            _receiveTh = new Thread(new ThreadStart(ReceiveThread));
            _receiveTh.Start();
            _log = new OperationLog();
            Clock = new System.Threading.Timer(ClockTick, null, 30000, 30000);
        }

        private void ReceiveThread()
        {
            NetworkStream ns = Client.GetStream();
            StreamReader reader = new StreamReader(ns, new UTF8Encoding());
            string line;
            while (Client.Connected)
            {
                try
                {
                    line = reader.ReadLine();
                    if (line == null)
                    {
                        Disconnect();
                        continue;
                    }
                    NetworkOperation op = null;
                    if (line.Substring(0, 3) != "MSG")
                    {
                        op = new NetworkCommand(line);
                    }
                    else
                    {
                        op = new NetworkMessage(line);
                        int numlineas = Convert.ToInt32(op.GetParameter(op.CountParameters() - 1));

                        for (int c = 0; c < numlineas; c++)
                        {
                            string datal = reader.ReadLine();
                            ((NetworkMessage)op).AddData(datal);
                        }
                    }
                    if (_log.IsWaiting(op))
                    {
                        _log.SendSignal(op, true);
                    }
                    else if (op.Name == "PNG")
                    {
                        op.AddParameter("OK", false);
                        op.Name = "QNG";
                        Send(op, false);
                    }
                    else if (op.Name != "QNG")
                    {
                        if (OnReceive != null)
                            OnReceive(Id, op);
                    }
                }
                catch
                {

                }
            }
            reader.Close();
            if (OnClientDisconnect != null) 
                OnClientDisconnect(Id);
        }

        public NetworkOperation Send(NetworkOperation op, bool wait)
        {
            if (Client.Connected)
            {
                op.trID = _trId;
                _trId++;

                LogEntry entry = null;
                if (wait) entry = _log.Add(op);

                NetworkStream ns = Client.GetStream();
                UTF8Encoding encoder = new UTF8Encoding();
                StringCollection opdata = op.toStreamDlg();

                foreach (string line in opdata)
                {
                    byte[] bytes = encoder.GetBytes(line + "\n");
                    ns.Write(bytes, 0, bytes.Length);
                }

                encoder = null;
                if (wait)
                {
                    entry.Locking.WaitOne();
                    return entry.Response;
                }
            }
            return null;            
        }

        public void Disconnect()
        {
            if (_log != null)
            {
                foreach (LogEntry entry in _log)
                    entry.Locking.Set();
                _log.Clear();
            }

            Client.Close();
           
            OnClientDisconnect(Id);
            Clock = null;
        }

        private void ClockTick(object state)
        {
            NetworkCommand PNG = new NetworkCommand("PNG", 0);
            Send(PNG, false);
        }
    }
}